from typing import Tuple

from django.core.exceptions import ValidationError


class ApiFormException(Exception):
    """Generic Django API Form exception"""
    pass


class UnsupportedMediaType(ApiFormException):
    """Unable to parse the request (based on the Content-Type)"""
    pass


class DetailValidationError(ValidationError):
    def __init__(self, error: ValidationError, path: Tuple):
        if not hasattr(error, 'message') and isinstance(error.error_list, list):
            for item in error.error_list:
                item.path = path

        super().__init__(error)
        self._path = path

    @property
    def path(self) -> Tuple:
        return self._path

    def prepend(self, key: Tuple):
        self._path = key + self._path

    def to_list(self) -> list:
        return list(self.path)

    def to_dict(self) -> dict:
        return {
            'code': self.code,
            'message': self.message,
            'path': self.to_list()
        }



import copy
from typing import List, Tuple

from django.core.exceptions import ValidationError
from django.forms import fields_for_model
from django.forms.forms import DeclarativeFieldsMetaclass as DjangoDeclarativeFieldsMetaclass
from django.forms.models import ModelFormOptions
from django.utils.translation import gettext as _

from .exceptions import UnsupportedMediaType, ApiFormException, DetailValidationError
from .settings import Settings
from .utils import resolve_from_path


class BaseForm:
    def __init__(self, data=None, request=None, settings: Settings = None):
        if data is None:
            self._data = {}
        else:
            self._data = data
        self.fields = copy.deepcopy(getattr(self, 'base_fields'))
        self._errors = None
        self._dirty = []
        self.cleaned_data = None
        self._request = request
        self.settings = settings or Settings()

        if isinstance(self.Meta, type):
            if hasattr(self.Meta, 'field_type_strategy'):
                for key in self.Meta.field_type_strategy.keys():
                    self.settings.POPULATION_STRATEGIES[key] = self.Meta.field_type_strategy[key]
            if hasattr(self.Meta, 'mapping'):
                for key in data.copy():
                    if key in self.Meta.mapping.keys():
                        data[self.Meta.mapping[key]] = data.pop(key)

        if isinstance(data, dict):
            for key in data.keys():
                if key in self.fields.keys():
                    self._dirty.append(key)

    def __getitem__(self, name):
        try:
            field = self.fields[name]
        except KeyError:
            raise KeyError(
                "Key '{}' not found in '{}'. Choices are: {}.".format(
                    name,
                    self.__class__.__name__,
                    ', '.join(sorted(self.fields)),
                )
            )
        return field

    def __iter__(self):
        for name in self.fields:
            yield self[name]

    @classmethod
    def create_from_request(cls, request):
        """
        :rtype: BaseForm
        """
        if not request.body:
            return cls()

        settings = Settings()

        all_attributes = request.META.get('CONTENT_TYPE', '').replace(' ', '').split(';')
        content_type = all_attributes.pop(0)

        optional_attributes = {}
        for attribute in all_attributes:
            key, value = attribute.split('=')
            optional_attributes[key] = value

        if content_type not in settings.PARSERS:
            raise UnsupportedMediaType()

        parser = resolve_from_path(settings.PARSERS[content_type])
        data = parser(request.body)

        return cls(data, request, settings)

    @property
    def dirty(self) -> List:
        return self._dirty

    @property
    def errors(self) -> dict:
        if not self._errors:
            self.full_clean()
        return self._errors

    def is_valid(self) -> bool:
        return not self.errors

    def add_error(self, field: Tuple, errors: ValidationError):
        if hasattr(errors, 'error_dict'):
            for key, items in errors.error_dict.items():
                for error in items:
                    if isinstance(error, DetailValidationError):
                        error.prepend(field)
                        self.add_error(error.path, error)
                    elif isinstance(error, ValidationError):
                        self.add_error(field + (key, ), error)
        elif not hasattr(errors, 'message') and isinstance(errors.error_list, list):
            for item in errors.error_list:
                if isinstance(item, DetailValidationError):
                    item.prepend(field)
                    self.add_error(item.path, item)
                elif isinstance(item, ValidationError):
                    path = field
                    if hasattr(item, 'path'):
                        path = field + item.path
                    self.add_error(path, item)
        else:
            self._errors.append(
                DetailValidationError(errors, (field,) if isinstance(field, str) else field)
            )

            if field in self.cleaned_data:
                del self.cleaned_data[field]

    def full_clean(self):
        """
        Clean all of self.data and populate self._errors and self.cleaned_data.
        """
        self._errors = []
        self.cleaned_data = {}

        for key, field in self.fields.items():
            try:
                if key in self.dirty or field.required:
                    validated_form_item = field.clean(self._data.get(key, None))

                    self.cleaned_data[key] = validated_form_item

                    if hasattr(self, f"clean_{key}"):
                        self.cleaned_data[key] = getattr(self, f"clean_{key}")()
            except ValidationError as e:
                self.add_error((key, ), e)
            except (AttributeError, TypeError, ValueError):
                self.add_error((key, ), ValidationError(_("Invalid value")))
        try:
            cleaned_data = self.clean()
        except ValidationError as e:
            self.add_error(('$body', ), e)
        else:
            if cleaned_data is not None:
                self.cleaned_data = cleaned_data

    def clean(self):
        """
        Hook for doing any extra form-wide cleaning after Field.clean() has been
        called on every field. Any ValidationError raised by this method will
        not be associated with a particular field; it will have a special-case
        association with the field named '$body'.
        """
        return self.cleaned_data

    def populate(self, obj, exclude: List[str] = None):
        """
                :param exclude:
                :param obj:
                :return:
                """
        if exclude is None:
            exclude = []

        if self.cleaned_data is None:
            raise ApiFormException("No clean data provided! Try to call is_valid() first.")

        for key, field in self.fields.items():
            # Skip if field is in exclude
            if key in exclude:
                continue

            # Skip if field is not in validated data
            if key not in self.cleaned_data.keys():
                continue

            field_class = f"{field.__class__.__module__}.{field.__class__.__name__}"
            strategy = resolve_from_path(
                self.settings.POPULATION_STRATEGIES.get(
                    field_class, "django_api_forms.population_strategies.BaseStrategy"
                )
            )
            if isinstance(self.Meta, type):
                if hasattr(self.Meta, 'field_strategy'):
                    if key in self.Meta.field_strategy.keys():
                        strategy = resolve_from_path(
                            self.Meta.field_strategy[key]
                        )

            strategy()(field, obj, key, self.cleaned_data[key])

        return obj


class DeclarativeFieldsMetaclass(DjangoDeclarativeFieldsMetaclass):
    """Collect Fields declared on the base classes."""
    def __new__(mcs, name, bases, attrs):
        new_class = super().__new__(mcs, name, bases, attrs)

        new_class.Meta = attrs.pop('Meta', None)

        return new_class


class ModelForm(BaseForm, metaclass=DeclarativeFieldsMetaclass):
    """
    SUPER EXPERIMENTAL
    """
    def __new__(cls, *args, **kwargs):
        new_class = super().__new__(cls, **kwargs)
        config = getattr(cls, 'Meta', None)

        model_opts = ModelFormOptions(getattr(config.model, '_meta', None))
        model_opts.exclude = getattr(config, 'exclude', tuple())

        fields = fields_for_model(
            model=model_opts.model,
            fields=None,
            exclude=model_opts.exclude,
            widgets=None,
            formfield_callback=None,
            localized_fields=model_opts.localized_fields,
            labels=model_opts.labels,
            help_texts=model_opts.help_texts,
            error_messages=model_opts.error_messages,
            field_classes=model_opts.field_classes,
            apply_limit_choices_to=False,
        )

        # AutoField has to be added manually
        # Do not add AutoFields right now
        # if config.model._meta.auto_field and config.model._meta.auto_field.attname not in model_opts.exclude:
        #     fields[config.model._meta.auto_field.attname] = IntegerField()

        fields.update(new_class.declared_fields)
        # Remove None value keys
        fields = {k: v for k, v in fields.items() if v is not None}

        new_class.base_fields = fields
        new_class.declared_fields = fields

        return new_class


class Form(BaseForm, metaclass=DeclarativeFieldsMetaclass):
    """A collection of Fields, plus their associated data."""



from importlib import import_module


def resolve_from_path(path: str):
    module_path, class_name = path.rsplit('.', 1)
    module = import_module(module_path)
    return getattr(module, class_name)



import typing
import warnings
from base64 import b64decode
from enum import Enum
from io import BytesIO
from mimetypes import guess_type
import re

from django.core.exceptions import ValidationError
from django.core.files import File
from django.forms import Field
from django.utils.translation import gettext_lazy as _

from .exceptions import DetailValidationError, ApiFormException
from .version import __version__ as version

DATA_URI_PATTERN = r"data:((?:\w+\/(?:(?!;).)+)?)((?:;[\w=]*[^;])*),(.+)"


class BooleanField(Field):
    def to_python(self, value):
        if value in (True, 'True', 'true', '1', 1):
            return True
        elif value in (False, 'False', 'false', '0', 0):
            return False
        else:
            return None

    def validate(self, value):
        if value is None and self.required:
            raise ValidationError(self.error_messages['required'], code='required')

    def has_changed(self, initial, data):
        if self.disabled:
            return False

        return self.to_python(initial) != self.to_python(data)


class FieldList(Field):
    default_error_messages = {
        'max_length': _('Ensure this list has at most %(max)d values (it has %(length)d).'),
        'min_length': _('Ensure this list has at least %(min)d values (it has %(length)d).'),
        'not_field': _('Invalid Field type passed into FieldList!'),
        'not_list': _('This field needs to be a list of objects!'),
    }

    def __init__(self, field, min_length=None, max_length=None, **kwargs):
        super().__init__(**kwargs)

        if not isinstance(field, Field):
            raise ApiFormException(self.error_messages['not_field'])

        self._min_length = min_length
        self._max_length = max_length
        self._field = field

    def to_python(self, value) -> typing.List:
        if not value:
            return []

        if not isinstance(value, list):
            raise ValidationError(self.error_messages['not_list'], code='not_list')

        if self._min_length is not None and len(value) < self._min_length:
            params = {'min': self._min_length, 'length': len(value)}
            raise ValidationError(self.error_messages['min_length'], code='min_length', params=params)

        if self._max_length is not None and len(value) > self._max_length:
            params = {'max': self._max_length, 'length': len(value)}
            raise ValidationError(self.error_messages['max_length'], code='max_length', params=params)

        result = []
        errors = []

        for position, item in enumerate(value):
            try:
                result.append(self._field.clean(item))
            except ValidationError as e:
                errors.append(DetailValidationError(e, (position,)))

        if errors:
            raise ValidationError(errors)

        return result


class FormField(Field):
    def __init__(self, form: typing.Type, **kwargs):
        self._form = form

        super().__init__(**kwargs)

    @property
    def form(self):
        return self._form

    def to_python(self, value) -> typing.Union[typing.Dict, None]:
        if not value:
            return {}

        form = self._form(value)
        if form.is_valid():
            return form.cleaned_data
        else:
            raise ValidationError(form.errors)


class FormFieldList(FormField):
    def __init__(self, form: typing.Type, min_length=None, max_length=None, **kwargs):
        self._min_length = min_length
        self._max_length = max_length
        super().__init__(form, **kwargs)

    default_error_messages = {
        'max_length': _('Ensure this list has at most %(max)d values (it has %(length)d).'),
        'min_length': _('Ensure this list has at least %(min)d values (it has %(length)d).'),
        'not_list': _('This field needs to be a list of objects!')
    }

    def to_python(self, value):
        if not value:
            return []

        if not isinstance(value, list):
            raise ValidationError(self.error_messages['not_list'], code='not_list')

        if self._min_length is not None and len(value) < self._min_length:
            params = {'min': self._min_length, 'length': len(value)}
            raise ValidationError(self.error_messages['min_length'], code='min_length', params=params)

        if self._max_length is not None and len(value) > self._max_length:
            params = {'max': self._max_length, 'length': len(value)}
            raise ValidationError(self.error_messages['max_length'], code='max_length', params=params)

        result = []
        errors = []

        for position, item in enumerate(value):
            form = self._form(item)
            if form.is_valid():
                result.append(form.cleaned_data)
            else:
                for error in form.errors:
                    error.prepend((position, ))
                    errors.append(error)

        if errors:
            raise ValidationError(errors)

        return result


class EnumField(Field):
    default_error_messages = {
        'not_enum': _('Invalid Enum type passed into EnumField!'),
        'invalid': _('Invalid enum value "{}" passed to {}'),
    }

    def __init__(self, enum: typing.Type, **kwargs):
        super().__init__(**kwargs)

        # isinstance(enum, type) prevents "TypeError: issubclass() arg 1 must be a class"
        # based on: https://github.com/samuelcolvin/pydantic/blob/v0.32.x/pydantic/utils.py#L260-L261
        if not (isinstance(enum, type) and issubclass(enum, Enum)):
            raise ApiFormException(self.error_messages['not_enum'])

        self.enum = enum

    def to_python(self, value) -> typing.Union[typing.Type[Enum], None]:
        if value is not None:
            try:
                return self.enum(value)
            except ValueError:
                raise ValidationError(self.error_messages['invalid'].format(value, self.enum), code='invalid')
        return None


class DictionaryField(Field):
    default_error_messages = {
        'not_field': _('Invalid Field type passed into DictionaryField!'),
        'not_dict': _('Invalid value passed to DictionaryField (got {}, expected dict)'),
    }

    def __init__(self, *, value_field, key_field=None, **kwargs):
        super().__init__(**kwargs)

        if not isinstance(value_field, Field):
            raise ApiFormException(self.error_messages['not_field'])

        if key_field and not isinstance(key_field, Field):
            raise ApiFormException(self.error_messages['not_field'])

        self._value_field = value_field
        self._key_field = key_field

    def to_python(self, value) -> dict:
        if not isinstance(value, dict):
            msg = self.error_messages['not_dict'].format(type(value))
            raise ValidationError(msg)

        result = {}
        errors = {}

        for key, item in value.items():
            try:
                if self._key_field:
                    key = self._key_field.clean(key)
                result[key] = self._value_field.clean(item)
            except ValidationError as e:
                errors[key] = DetailValidationError(e, (key, ))

        if errors:
            raise ValidationError(errors)

        return result


class AnyField(Field):
    def to_python(self, value) -> typing.Union[typing.Dict, typing.List]:
        return value


class FileField(Field):
    default_error_messages = {
        'max_length': _('Ensure this file has at most %(max)d bytes (it has %(length)d).'),
        'invalid_uri': _("The given URI is not a valid Data URI."),
        'invalid_mime': _("The submitted file is empty."),
    }

    def __init__(self, max_length=None, mime: typing.Tuple = None, **kwargs):
        self._max_length = max_length
        self._mime = mime
        super().__init__(**kwargs)

    def to_python(self, value: str) -> typing.Optional[File]:
        if not value:
            return None

        if re.fullmatch(DATA_URI_PATTERN, value) is None:
            warnings.warn(
                "Raw base64 inside of FileField/ImageField is deprecated and will throw a validation error from "
                "version >=1.0.0 Provide value as a Data URI.",
                DeprecationWarning
            )
            if version >= "1.0.0":
                raise ValidationError(self.error_messages["invalid_uri"], code="invalid_uri")

        mime = None

        if ',' in value:
            mime, strict = guess_type(value)
            value = value.split(',')[-1]

        if self._mime and mime not in self._mime:
            params = {'allowed': ', '.join(self._mime), 'received': mime}
            raise ValidationError(self.error_messages['invalid_mime'], code='invalid_mime', params=params)

        file = File(BytesIO(b64decode(value)))

        if self._max_length is not None and file.size > self._max_length:
            params = {'max': self._max_length, 'length': file.size}
            raise ValidationError(self.error_messages['max_length'], code='max_length', params=params)

        file.content_type = mime

        return file


class ImageField(FileField):
    default_error_messages = {
        'invalid_image': _("Upload a valid image. The file you uploaded was either not an image or a corrupted image.")
    }

    def to_python(self, value) -> typing.Optional[File]:
        f = super().to_python(value)

        if f is None:
            return None

        # Pillow is required for ImageField
        from PIL import Image

        file = BytesIO(f.read())  # Create fp for Pillow

        try:
            image = Image.open(file)
            image.verify()
            f.image = Image.open(file)  # Image have to be reopened after Image.verify() call
            f.content_type = Image.MIME.get(image.format)
        except Exception:
            raise ValidationError(
                self.error_messages['invalid_image'],
                code='invalid_image'
            )

        if self._mime and f.content_type not in self._mime:
            params = {'allowed': ', '.join(self._mime), 'received': f.content_type}
            raise ValidationError(self.error_messages['invalid_mime'], code='invalid_mime', params=params)

        f.seek(0)  # Return to start of the file

        return f



from django.conf import settings

DEFAULTS = {
    'POPULATION_STRATEGIES': {
        'django_api_forms.fields.FormFieldList': 'django_api_forms.population_strategies.IgnoreStrategy',
        'django_api_forms.fields.FileField': 'django_api_forms.population_strategies.IgnoreStrategy',
        'django_api_forms.fields.ImageField': 'django_api_forms.population_strategies.IgnoreStrategy',
        'django_api_forms.fields.FormField': 'django_api_forms.population_strategies.IgnoreStrategy',
        'django.forms.models.ModelMultipleChoiceField': 'django_api_forms.population_strategies.IgnoreStrategy',
        'django.forms.models.ModelChoiceField': 'django_api_forms.population_strategies.ModelChoiceFieldStrategy'
    },
    'DEFAULT_POPULATION_STRATEGY': 'django_api_forms.population_strategies.BaseStrategy',
    'PARSERS': {
        'application/json': 'json.loads',
        'application/x-msgpack': 'msgpack.loads'
    }
}


class Settings:
    def __getattr__(self, item):
        if item not in DEFAULTS:
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{item}'")

        django_setting = f"DJANGO_API_FORMS_{item}"
        default = DEFAULTS[item]

        if hasattr(settings, django_setting):
            customized_value = getattr(settings, django_setting)
            if isinstance(default, dict):
                value = {**default, **customized_value}
            else:
                value = customized_value
        else:
            value = default

        setattr(self, item, value)
        return value



class BaseStrategy:
    def __call__(self, field, obj, key: str, value):
        setattr(obj, key, value)


class IgnoreStrategy(BaseStrategy):
    def __call__(self, field, obj, key: str, value):
        pass


class ModelChoiceFieldStrategy(BaseStrategy):

    """
    We need to change key postfix if there is a ModelChoiceField (because of _id etc.)
    We always try to assign whole object instance, for example:
    artis_id is normalized as Artist model, but it have to be assigned to artist model property
    because artist_id in model has different type (for example int if your are using int primary keys)
    If you are still confused (sorry), try to check docs
    """
    def __call__(self, field, obj, key: str, value):
        model_key = key
        if field.to_field_name:
            postfix_to_remove = f"_{field.to_field_name}"
        else:
            postfix_to_remove = "_id"
        if key.endswith(postfix_to_remove):
            model_key = key[:-len(postfix_to_remove)]
        setattr(obj, model_key, value)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['django_api_forms']

package_data = \
{'': ['*']}

install_requires = \
['Django>=2.0', 'Pillow>=2.1', 'msgpack']

setup_kwargs = {
    'name': 'django-api-forms',
    'version': '1.0.0rc5',
    'description': 'Declarative Django request validation for RESTful APIs',
    'long_description': 'None',
    'author': 'Jakub Dubec',
    'author_email': 'jakub.dubec@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from .exceptions import DetailValidationError
from .fields import BooleanField
from .fields import FieldList
from .fields import FormField
from .fields import FormFieldList
from .fields import EnumField
from .fields import DictionaryField
from .fields import AnyField
from .fields import FileField
from .fields import ImageField
from .forms import Form
from .forms import ModelForm
from .version import __version__

__all__ = [
    'DetailValidationError',
    'BooleanField',
    'FieldList',
    'FormField',
    'FormFieldList',
    'EnumField',
    'DictionaryField',
    'AnyField',
    'FileField',
    'ImageField',
    'Form',
    'ModelForm',
    '__version__'
]
